module Parser (
  parseFOL
) where

import Text.Parsec
import Text.Parsec.String
--import Text.Parsec.Expr
import Text.Parsec.Token
import Text.Parsec.Language

import FOLData

parseFOL :: String -> Formula
parseFOL s =
    case (parse fullFormula "FOLParse" s) of
      Left e -> error $ show e
      Right f -> f

fullFormula :: Parser Formula
fullFormula = do
  f <- formula
  eof
  return f

-- Sentence parsers

formula :: Parser Formula
formula = sQuant <|> sConn <|> sNot <|> sAtomic

sAtomic :: Parser Formula
sAtomic = sPred <|> sBracket

sBracket :: Parser Formula
sBracket = mParens formula

sPred :: Parser Formula
sPred = do 
  p <- predicate
  ts <- (mParens terms) <|> (return [])
  return (Pred p ts)


sConn :: Parser Formula
sConn = let
    s1 = sAtomic <|> sNot
    s2 = sAnd <|> s1
    s3 = sOr <|> s2
    --s4 = sImply <|> s3
    sAnd = s1 `chainl1` cAnd
    sOr = s2 `chainl1` cOr
    sImply = s3 `chainl1` cImply
  in
    sImply <|> sOr <|> sAnd

cImply :: Parser (Formula -> Formula -> Formula)
cImply = do 
  cimp
  return (\s1 -> \s2 -> Conn s1 Imp s2)

cAnd :: Parser (Formula -> Formula -> Formula)
cAnd = do 
  cand
  return (\s1 -> \s2 -> Conn s1 And s2)

cOr :: Parser (Formula -> Formula -> Formula)
cOr = do
  cor
  return (\s1 -> \s2 -> Conn s1 Or s2)

sQuant :: Parser Formula
sQuant = do
  q <- quantifier
  vs <- variable
  mReservedOp "."
  s <- formula
  return (Quant q vs s)

sNot :: Parser Formula
sNot = do 
  mReservedOp "NOT" <|> mReservedOp "Not" <|> mReservedOp "not" <|> mReservedOp "~"
  s <- sAtomic
  return (Not s)

-- Term parsers

term :: Parser Term
term = try tFunc <|> tConst <|> tVar

terms :: Parser [Term]
terms = term `sepBy1` char ','

tFunc :: Parser Term
tFunc = do
  f <- function
  ts <- mParens terms
  return (Func f ts)

tConst :: Parser Term
tConst = do
  c <- constant
  return (Const c)

tVar :: Parser Term
tVar = do
  v <- variable
  return (Var v)

-- Connective parser

--connective :: Parser Connective
--connective = cand <|> cor <|> cimp

cand :: Parser Connective
cand = do
  mReservedOp "AND" <|> mReservedOp "And" <|> mReservedOp "and" <|> mReservedOp "/\\"
  return And
  
cor :: Parser Connective
cor = do 
  mReservedOp "OR" <|> mReservedOp "Or" <|> mReservedOp "or" <|> mReservedOp "\\/"
  return Or

cimp :: Parser Connective
cimp = do
  mReservedOp "=>" <|> mReservedOp "==>" <|> mReservedOp "imp" <|> mReservedOp "->"
  return Imp

-- Quatifier parser

quantifier :: Parser Quantifier
quantifier = exists <|> forall

forall :: Parser Quantifier
forall = do 
  mReservedOp "FORALL" <|> mReservedOp "ForAll" <|> mReservedOp "forall" <|> mReservedOp "!"
  return ForAll
  
exists :: Parser Quantifier
exists = do 
  mReservedOp "EXISTS" <|> mReservedOp "Exists" <|> mReservedOp "exists" <|> mReservedOp "?"
  return Exists

-- Lexers

               
keywords :: [String]
keywords = [ "TRUE", "True", "true",
       "FALSE", "False", "false",
       "NOT", "Not", "not", "~",
       "AND", "And", "and" , "/\\",
       "OR", "Or", "or", "\\/",
       "FORALL", "ForAll", "forall","!",
       "EXISTS", "Exists", "exists","?",
             ".","imp","->"]

variable :: Parser String
variable = mIdentifier

--variables :: Parser [String]
--variables = variable `sepBy1` char ','

constant, predicate, function :: Parser String
constant = cIdentifier
predicate = cIdentifier
function = cIdentifier


def :: LanguageDef st
def = emptyDef{ commentStart = "{-"
              , commentEnd = "-}"
              , identStart = letter
              , identLetter = alphaNum
              , reservedOpNames = keywords
              }


mIdentifier :: Parser String
mReservedOp :: String -> Parser ()
--mWhiteSpace :: Parser ()
TokenParser{ parens = mParens
           , identifier = mIdentifier
           , reservedOp = mReservedOp
           --, whiteSpace = mWhiteSpace
           } = makeTokenParser def

defCap :: LanguageDef st
defCap = emptyDef{ commentStart = "{-"
              , commentEnd = "-}"
              , identStart = upper
              , identLetter = alphaNum
              , reservedOpNames = keywords
              }

cIdentifier:: Parser String
TokenParser{
  identifier = cIdentifier } = makeTokenParser defCap


