module TPTPtoFOL 
( TPTPFormat(..)
, parseTPTP
, parseTPTPFile
) where

import Data.Functor.Identity
import Codec.TPTP
import qualified FOLData as F

data TPTPFormat = CNFFormat | FOFFormat

parseTPTPFile :: TPTPFormat -> String -> IO F.Formula
parseTPTPFile fmt file = do
  inp <- parseFile file
  return (convertTPTPToFOL fmt inp)

parseTPTP :: TPTPFormat -> String -> F.Formula
parseTPTP fmt s = let
  inp = parse s
  in convertTPTPToFOL fmt inp

convertTPTPToFOL :: TPTPFormat -> [TPTP_Input] -> F.Formula 
convertTPTPToFOL FOFFormat inp = convertTPTPFOFToFOL inp
convertTPTPToFOL CNFFormat inp = convertTPTPCNFToFOL inp

convertTPTPCNFToFOL :: [TPTP_Input] -> F.Formula
convertTPTPCNFToFOL tptpIn = 
  if null nconjectures
  then ax
  else F.Conn ax F.And (ncj)
  where
    mbfs = map processInput tptpIn
    axioms = filter (condRole ["axiom","hypothesis"]) mbfs
    nconjectures = filter (condRole ["negated_conjecture"]) mbfs
    condRole xs (Just (x,_))= unrole x `elem` xs
    condRole _ Nothing = False
    getFormula (Just x) = snd x
    getFormula Nothing = error $ errorMsg "conversion returned a non formula."
    ax = conjunctOfFormulas (map getFormula axioms)
    ncj = conjunctOfFormulas (map getFormula nconjectures)
    
convertTPTPFOFToFOL :: [TPTP_Input] -> F.Formula
convertTPTPFOFToFOL tptpIn = F.Not (F.Conn ax F.Imp (cj))
  where
    mbfs = map processInput tptpIn
    axioms = filter (condRole ["axiom"]) mbfs
    conjectures = filter (condRole ["conjecture"]) mbfs
    condRole xs (Just (x,_))= unrole x `elem` xs
    condRole _ Nothing = False
    getFormula (Just x) = snd x
    getFormula Nothing = error $ errorMsg "conversion returned a non formula."
    ax = conjunctOfFormulas (map getFormula axioms)
    cj = conjunctOfFormulas (map getFormula conjectures)
    

processInput :: TPTP_Input -> Maybe (Role,F.Formula)
processInput (Comment _) = Nothing
processInput (Include _ _) = Nothing
processInput (AFormula _ r f _) = Just (r,tpFormulaToFormula f)


tpFormulaToFormula :: Formula -> F.Formula
tpFormulaToFormula (F f) = tpFormulaToFormula1 (runIdentity f)


tpFormulaToFormula1 :: Formula0 Term Formula -> F.Formula 
tpFormulaToFormula1 (Quant q vs f) = foldr (F.Quant q') (F.Quant q' (last vs') f') (init vs')
  where
    q' = case q of
      Exists -> F.Exists
      All -> F.ForAll
    f' = tpFormulaToFormula f
    vs' = map variableToString vs
tpFormulaToFormula1 (PredApp p ts) = F.Pred p' ts' 
  where
    p' = atomicWordtoString p
    ts' = map tpTermToTerm ts
tpFormulaToFormula1 ((:~:) f) = F.Not (tpFormulaToFormula f)
tpFormulaToFormula1 (BinOp f1 op f2) =
  case op of
    (:<=>:) -> F.Conn (F.Conn f1' F.Imp f2') F.And (F.Conn f2' F.Imp f1')
    (:=>:)  -> F.Conn f1' F.Imp f2' 
    (:<=:)  -> F.Conn f2' F.Imp f1'  
    (:&:)   -> F.Conn f1' F.And f2'
    (:|:)   -> F.Conn f1' F.Or  f2'
    (:~&:)  -> F.Not (F.Conn f1' F.And f2')
    (:~|:)  -> F.Not (F.Conn f1' F.Or f2')
    (:<~>:) -> F.Conn (F.Conn (F.Not f1') F.And f2') F.Or (F.Conn f2' F.And (F.Not f1'))
  where
    f1' = tpFormulaToFormula f1
    f2' = tpFormulaToFormula f2
tpFormulaToFormula1 (InfixPred t1 (:=:) t2) = F.Pred F.equalPredicate [t1',t2']
  where
    t1' = tpTermToTerm t1
    t2' = tpTermToTerm t2
tpFormulaToFormula1 (InfixPred t1 (:!=:) t2) = F.Not (F.Pred F.equalPredicate [t1',t2'])
  where
    t1' = tpTermToTerm t1
    t2' = tpTermToTerm t2
    
tpTermToTerm :: Term -> F.Term
tpTermToTerm (T t) = tpTerm0ToTerm (runIdentity t)

tpTerm0ToTerm :: Term0 Term -> F.Term
tpTerm0ToTerm (Var v) = F.Var (variableToString v)
tpTerm0ToTerm (NumberLitTerm _) = error $ errorMsg "number literals not supported."
tpTerm0ToTerm (DistinctObjectTerm s) = F.Const s
tpTerm0ToTerm (FunApp f ts) = 
  if null ts
  then F.Const f'
  else F.Func f' ts'
  where
    f' = atomicWordtoString f
    ts' = map tpTermToTerm ts 


variableToString :: V -> String
variableToString (V s) = s

atomicWordtoString :: AtomicWord -> String
atomicWordtoString (AtomicWord s) = s

conjunctOfFormulas :: [F.Formula] -> F.Formula
conjunctOfFormulas fs = foldr (\f cf -> F.Conn f F.And cf) (last fs) (init fs)

errorMsg :: String -> String
errorMsg s = "TPTP to FOL conversion: "++s
